import sys

if sys.version_info[0] == 2:
    print("Fatal: this application requires Python3. Please run it in a virtualenv to make sure \"python --version\" gives 3.x.y.")
    sys.exit(2)

from bumpsemver.cli import main

main()



# pylint: skip-file

# noinspection PyPackageRequirements
import pytest
# noinspection PyPackageRequirements
from testfixtures import LogCapture

from textwrap import dedent
from ruamel.yaml import YAML

from bumpsemver import exceptions
from bumpsemver.cli import main


@pytest.fixture(params=[
    "yaml",
    "yaml(suffix)",
    "yaml (suffix with space)",
    "yaml (suffix lacking closing paren",
])
def yaml_keyword(request):
    """Return multiple possible styles for the bumpsemver:yaml keyword."""
    return request.param


def test_yaml_file_simple(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 85.0.1
        [bumpsemver:yaml:playbook.yml]
        yamlpath = *.vars.project_version
        """).strip())
    tmpdir.join("playbook.yml").write(dedent("""
        ---
        - name: 'create CodeBuild projects'
          vars:
            project_version: '85.0.1'
            software_component: 'devops'
    """).strip())
    main(["minor"])
    assert "current_version = 85.1.0" in tmpdir.join(".bumpsemver.cfg").read()
    yaml = YAML()
    data = yaml.load(tmpdir.join("playbook.yml").read())
    assert data[0]['vars']['project_version'] == "85.1.0"


def test_yaml_file_exact_path(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 85.2.1
        [bumpsemver:yaml:playbook.yml]
        yamlpath = *.vars.project_version
        """).strip())
    tmpdir.join("playbook.yml").write(dedent("""
        - level: '[]-prop'
          vars:
            project_version: 85.2.1
          child_prop1:
            level: '[]-prop-prop'
            vars:
              project_version: 85.2.1
          child_prop2:
            - level: '[]-prop-[]-prop'
              vars:
                project_version: 85.2.1
            """).strip())
    main(["minor"])
    assert "current_version = 85.3.0" in tmpdir.join(".bumpsemver.cfg").read()
    yaml = YAML()
    data = yaml.load(tmpdir.join("playbook.yml").read())
    assert data[0]['vars']['project_version'] == "85.3.0"
    assert data[0]['child_prop1']['vars']['project_version'] == "85.2.1"
    assert data[0]['child_prop2'][0]['vars']['project_version'] == "85.2.1"


def test_yaml_file_with_suffix_two(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 85.4.1
        [bumpsemver:yaml(a):playbook.yml]
        yamlpath = version
        [bumpsemver:yaml(b):playbook.yml]
        yamlpath = pos.version
        """).strip())
    tmpdir.join("playbook.yml").write(dedent("""
        version: 85.4.1
        pos:
          version: 85.4.1
        neg:
          version: 85.4.1
    """).strip())
    main(["minor"])
    assert "current_version = 85.5.0" in tmpdir.join(".bumpsemver.cfg").read()
    yaml = YAML()
    data = yaml.load(tmpdir.join("playbook.yml").read())
    assert data['version'] == "85.5.0"
    assert data['pos']['version'] == "85.5.0"
    assert data['neg']['version'] == "85.4.1"


def test_yaml_file_with_suffix_complex(tmpdir, yaml_keyword):
    tmpdir.join("file2").write("version: 5.10.2")
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version: 5.10.2
        new_version: 5.10.8
        [bumpsemver:%s:file2]
        yamlpath: version
        """ % yaml_keyword).strip())
    tmpdir.chdir()
    main(["patch"])

    assert "5.10.8" in tmpdir.join("file2").read()


def test_yaml_file_key_does_not_exist(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 85.8.1
        [bumpsemver:yaml:playbook.yml]
        yamlpath = version_foo_bar
        """).strip())
    tmpdir.join("playbook.yml").write(dedent("""
        version: 85.8.1
    """).strip())

    with pytest.raises(exceptions.VersionNotFoundException,
                       match="Did not find '85.8.1' at yamlpath 'version_foo_bar' in file: 'playbook.yml'"):
        main(["minor"])

    assert "current_version = 85.8.1" in tmpdir.join(".bumpsemver.cfg").read()
    yaml = YAML()
    data = yaml.load(tmpdir.join("playbook.yml").read())
    assert data['version'] == "85.8.1"


def test_yaml_file_key_exists_but_wrong_value(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 85.10.1
        [bumpsemver:yaml:playbook.yml]
        yamlpath = version
        """).strip())
    tmpdir.join("playbook.yml").write(dedent("""
        version: ^85.10.1
    """).strip())

    with pytest.raises(exceptions.VersionNotFoundException, match="Did not find '85.10.1' at yamlpath 'version' in file: 'playbook.yml'"):
        main(["minor"])

    assert "current_version = 85.10.1" in tmpdir.join(".bumpsemver.cfg").read()
    yaml = YAML()
    data = yaml.load(tmpdir.join("playbook.yml").read())
    assert data['version'] == "^85.10.1"


def test_yaml_file_multiple_values_positive(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 85.11.1
        [bumpsemver:yaml:playbook.yml]
        yamlpath = *.version
        """).strip())
    tmpdir.join("playbook.yml").write(dedent("""
        ---
        - name: 'node 1'
          version: 85.11.1
        - name: 'node 2'
          version: 85.11.1
        - name: 'node 3'
          child:
            version: 85.11.1
    """).strip())
    main(["minor"])
    assert "current_version = 85.12.0" in tmpdir.join(".bumpsemver.cfg").read()
    yaml = YAML()
    data = yaml.load(tmpdir.join("playbook.yml").read())
    assert data[0]['version'] == "85.12.0"
    assert data[1]['version'] == "85.12.0"
    assert data[2]['child']['version'] == "85.11.1"


def test_yaml_file_multiple_values_negative(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 85.12.1
        [bumpsemver:yaml:playbook.yml]
        yamlpath = *.version
        """).strip())
    tmpdir.join("playbook.yml").write(dedent("""
        - name: 'node 1'
          version: 85.12.1
        - name: 'node 2'
          version: 1.2.3
    """).strip())

    with pytest.raises(exceptions.VersionNotFoundException,
                       match=r"Did not find '85.12.1' at yamlpath '\*\.version' in file: 'playbook.yml'"):
        main(["minor"])

    assert "current_version = 85.12.1" in tmpdir.join(".bumpsemver.cfg").read()
    yaml = YAML()
    data = yaml.load(tmpdir.join("playbook.yml").read())
    assert data[0]['version'] == "85.12.1"
    assert data[1]['version'] == "1.2.3"


def test_yaml_file_info_logging(tmpdir):
    tmpdir.join("fileY").write(dedent("""
        ---
        version: 0.4.0
        """).strip())
    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write(dedent(r"""
        [bumpsemver]
        current_version = 0.4.0

        [bumpsemver:yaml:fileY]
        yamlpath = version
        """).strip())

    with LogCapture() as log_capture:
        main(["patch", "--verbose"])

    log_capture.check(
        ("bumpsemver.cli", "INFO", "Reading config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO", "[bumpsemver]\ncurrent_version = 0.4.0\n\n[bumpsemver:yaml:fileY]\nyamlpath = version"),
        ("bumpsemver.version_part", "INFO", "Parsing version '0.4.0' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=0, minor=4, patch=0"),
        ("bumpsemver.cli", "INFO", "Attempting to increment part 'patch'"),
        ("bumpsemver.cli", "INFO", "Values are now: major=0, minor=4, patch=1"),
        ("bumpsemver.version_part", "INFO", "Parsing version '0.4.1' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=0, minor=4, patch=1"),
        ("bumpsemver.cli", "INFO", "New version will be '0.4.1'"),
        ("bumpsemver.cli", "INFO", "Asserting files fileY contain the version string..."),
        ("bumpsemver.files.yaml", "INFO", "Changing yaml file fileY:"),
        ("bumpsemver.files.yaml", "INFO",
         "--- a/fileY\n+++ b/fileY\n@@ -1,2 +1,2 @@\n ---\n-version: 0.4.0\n+version: 0.4.1"),
        ("bumpsemver.cli", "INFO", "Writing to config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO", "[bumpsemver]\ncurrent_version = 0.4.1\n\n[bumpsemver:yaml:fileY]\nyamlpath = version\n\n")
    )



import logging

from bumpsemver.version_part import Version

from bumpsemver.exceptions import VersionNotFoundException
from bumpsemver.files import FileTypes
from bumpsemver.files.base import FileTypeBase

logger = logging.getLogger(__name__)


class ConfiguredGenericFile(FileTypeBase):

    def __init__(self, path, version_config):
        self.path = path
        self._version_config = version_config
        self.file_type = FileTypes.GENERIC
        super().__init__(path, version_config, FileTypes.GENERIC, logger)

    def should_contain_version(self, version: Version, context: dict) -> None:
        """
        Raise VersionNotFound if the version number isn't present in this file.

        Return normally if the version number is in fact present.
        """
        context["current_version"] = self._version_config.serialize(version, context)
        search_expression = self._version_config.search.format(**context)

        if self.contains(search_expression):
            return

        # the `search` pattern did not match, but the original supplied version number (representing the same version part values) might
        # match instead.

        # check whether `search` isn't customized, i.e. should match only very specific parts of the file
        search_pattern_is_default = self._version_config.search == "{current_version}"

        if search_pattern_is_default and self.contains(version.original):
            # original version is present and we're not looking for something more specific -> this is accepted as a match
            return

        # version not found
        raise VersionNotFoundException(f"Did not find '{search_expression}' in file: '{self.path}'")

    def contains(self, search: str) -> bool:
        if not search:
            return False

        with open(self.path, "rt", encoding="utf-8") as orig_fp:
            search_lines = search.splitlines()
            lookbehind = []

            for lineno, line in enumerate(orig_fp.readlines()):
                lookbehind.append(line.rstrip("\n"))

                if len(lookbehind) > len(search_lines):
                    lookbehind = lookbehind[1:]

                if search_lines[0] in lookbehind[0] and search_lines[-1] in lookbehind[-1] and search_lines[1:-1] == lookbehind[1:-1]:
                    logger.info(f"Found '{search}' in {self.path} at line {lineno - (len(lookbehind) - 1)}: {line.rstrip()}")
                    return True
        return False

    def replace(self, current_version: Version, new_version: Version, context: dict, dry_run: bool) -> None:

        with open(self.path, "rt", encoding="utf-8") as orig_fp:
            file_content_before = orig_fp.read()

        context["current_version"] = self._version_config.serialize(current_version, context)
        context["new_version"] = self._version_config.serialize(new_version, context)

        search_for = self._version_config.search.format(**context)
        replace_with = self._version_config.replace.format(**context)

        file_content_after = file_content_before.replace(search_for, replace_with)

        if file_content_before == file_content_after:
            file_content_after = file_content_before.replace(current_version.original, replace_with)

        self.update_file(file_content_before, file_content_after, dry_run)

    def __repr__(self):
        return f"<bumpsemver.files.ConfiguredGenericFile:{self.path}>"



import logging
from types import SimpleNamespace

from ruamel.yaml.compat import StringIO
from yamlpath import Processor, YAMLPath
from yamlpath.common import Parsers
from yamlpath.exceptions import YAMLPathException
from yamlpath.wrappers import ConsolePrinter

from bumpsemver.exceptions import BumpVersionException, VersionNotFoundException
from bumpsemver.files import FileTypes
from bumpsemver.files.base import FileTypeBase
from bumpsemver.version_part import Version

logger = logging.getLogger(__name__)


class InvalidYAMLException(BumpVersionException):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class ConfiguredYAMLFile(FileTypeBase):
    def __init__(self, path, yamlpath, version_config):
        super().__init__(path, version_config, FileTypes.YAML, logger)
        self.yaml = Parsers.get_yaml_editor()
        self.yaml_path = yamlpath
        self.yaml_logging_args = SimpleNamespace(quiet=True, verbose=False, debug=False)
        self.yaml_log = ConsolePrinter(self.yaml_logging_args)

    def should_contain_version(self, version: Version, context: dict) -> None:
        current_version = self._version_config.serialize(version, context)
        context["current_version"] = current_version

        if self.contains(current_version):
            return

        # version not found
        raise VersionNotFoundException(f"Did not find '{current_version}' at yamlpath '{self.yaml_path}' in file: '{self.path}'")

    def contains(self, search: str) -> bool:
        try:
            processor = self.__get_processor()
            yaml_path = YAMLPath(self.yaml_path)
            nodes_found = list(processor.get_nodes(yaml_path, mustexist=True))
            matched = True
            for node in nodes_found:
                if node.node != search:
                    matched = False
            return matched
        except YAMLPathException as ex:
            logger.error(f"invalid path expression: {str(ex)}", exc_info=ex)
            return False

    def __dump(self, data) -> str:
        stream = StringIO()
        self.yaml.dump(data, stream)
        return stream.getvalue()

    def __get_processor(self):
        (yaml_data, doc_loaded) = Parsers.get_yaml_data(self.yaml, self.yaml_log, self.path)
        if not doc_loaded:
            raise InvalidYAMLException(f"Failed in reading YAML file '{self.path}'")
        processor = Processor(self.yaml_log, yaml_data)
        return processor

    def replace(self, current_version: Version, new_version: Version, context: dict, dry_run: bool) -> None:
        processor = self.__get_processor()
        file_content_before = self.__dump(processor.data)

        current_version = self._version_config.serialize(current_version, context)
        context["current_version"] = current_version
        new_version = self._version_config.serialize(new_version, context)
        context["new_version"] = new_version

        yaml_path = YAMLPath(self.yaml_path)
        for node in processor.get_nodes(yaml_path, mustexist=True):
            if node.node == current_version:
                processor.set_value(node.path, new_version)

        file_content_after = self.__dump(processor.data)

        self.update_file(file_content_before, file_content_after, dry_run)

    def __repr__(self):
        return f"<bumpsemver.ConfiguredYAMLFile:{self.path}>"



"""
Commonly used utilities
"""


def key_value_string(obj: dict) -> str:
    """
    Dump a dict object into a string representation of key-value pairs
    """
    return ", ".join(f"{k}={v}" for k, v in sorted(obj.items()))



import collections
import io
import json
import logging
from json import JSONDecodeError

from jsonpath_ng import parse

from bumpsemver.version_part import Version
from bumpsemver.exceptions import VersionNotFoundException
from bumpsemver.files import FileTypes
from bumpsemver.files.base import FileTypeBase

logger = logging.getLogger(__name__)


def _get_json_value(obj, path):
    json_path_expr = parse(path)
    match = json_path_expr.find(obj)
    return match[0].value


def _set_json_value(obj, path, value):
    json_path_expr = parse(path)
    json_path_expr.find(obj)
    json_path_expr.update(obj, value)


class ConfiguredJSONFile(FileTypeBase):
    def __init__(self, path, jsonpath, version_config):
        super().__init__(path, version_config, FileTypes.JSON, logger)
        self.json_path = jsonpath

    def should_contain_version(self, version: Version, context: dict) -> None:
        current_version = self._version_config.serialize(version, context)
        context["current_version"] = current_version

        if self.contains(current_version):
            return

        # version not found
        raise VersionNotFoundException(f"Did not find '{current_version}' at jsonpath '{self.json_path}' in file: '{self.path}'")

    def contains(self, search: str) -> bool:
        try:
            with io.open(self.path, "rt", encoding="utf-8") as orig_fp:
                data = json.load(orig_fp)
            return _get_json_value(data, self.json_path) == search
        except LookupError as err:
            logger.error(f"invalid path expression: {str(err)}", exc_info=err)
            return False
        except JSONDecodeError as err:
            raise err

    def replace(self, current_version: Version, new_version: Version, context: dict, dry_run: bool) -> None:
        with io.open(self.path, "rt", encoding="utf-8") as orig_fp:
            file_content_before = orig_fp.read()
            # the object_pairs_hook allows us to load the json in a way that key order is preserved and will keep the file diff
            # to a minimum
            #
            # noinspection PyTypeChecker
            data = json.loads(file_content_before, object_pairs_hook=collections.OrderedDict)

        current_version = self._version_config.serialize(current_version, context)
        context["current_version"] = current_version
        new_version = self._version_config.serialize(new_version, context)
        context["new_version"] = new_version

        if _get_json_value(data, self.json_path) == current_version:
            _set_json_value(data, self.json_path, new_version)
        # ensure_ascii: we're writing utf-8 files, so we don't need ascii support
        # allow_nan: JSON does not have an understanding of infinity or nan, so it’s forbidden
        # indent: indent of 2 spaces is common practise
        # separators: the default separators leave a trailing space after every comma. when using indentation, this results in awkward
        #             line-endings with a leading space
        file_content_after = json.dumps(data, ensure_ascii=False, allow_nan=False, indent=2, separators=(",", ": ")) + "\n"

        self.update_file(file_content_before, file_content_after, dry_run)

    def __repr__(self):
        return f"<bumpsemver.files.ConfiguredJSONFile:{self.path}>"



# pylint: skip-file

import argparse
import logging
import os
import subprocess
import warnings
from configparser import RawConfigParser
from datetime import datetime
from functools import partial
from shlex import split as shlex_split
from textwrap import dedent
from unittest import mock

# noinspection PyPackageRequirements
import pytest
# noinspection PyPackageRequirements
from testfixtures import LogCapture

import bumpsemver
from bumpsemver import exceptions
from bumpsemver.cli import DESCRIPTION, main, split_args_in_optional_and_positional

check_output = partial(subprocess.check_output, env=os.environ.copy())

COMMIT = r"""
[bumpsemver]
commit = True
""".strip()

COMMIT_NOT_TAG = r"""
[bumpsemver]
commit = True
tag = False
""".strip()


@pytest.fixture(params=[
    "file",
    "file(suffix)",
    "file (suffix with space)",
    "file (suffix lacking closing paren",
])
def file_keyword(request):
    """Return multiple possible styles for the bumpsemver:file keyword."""
    return request.param


RawConfigParser(empty_lines_in_values=False)


def _mock_calls_to_string(called_mock):
    return [f"{name}|{args[0] if len(args) > 0 else args}|{repr(kwargs) if len(kwargs) > 0 else ''}"
            for name, args, kwargs in called_mock.mock_calls]


EXPECTED_OPTIONS = r"""
[-h]
[--config-file FILE]
[--verbose]
[--allow-dirty]
[--search SEARCH]
[--replace REPLACE]
[--current-version VERSION]
[--dry-run]
--new-version VERSION
[--commit | --no-commit]
[--tag | --no-tag]
[--sign-tags | --no-sign-tags]
[--tag-name TAG_NAME]
[--tag-message TAG_MESSAGE]
[--message COMMIT_MSG]
part
[file ...]
""".strip().splitlines()

EXPECTED_USAGE = (r"""

%s

positional arguments:
  part                  Part of the version to be bumped.
  file                  Files to change (default: [])

optional arguments:
  -h, --help            show this help message and exit
  --config-file FILE    Config file to read most of the variables from
                        (default: .bumpsemver.cfg)
  --verbose             Print verbose logging to stderr (default: 0)
  --allow-dirty         Don't abort if working directory is dirty (default:
                        False)
  --search SEARCH       Template for complete string to search (default:
                        {current_version})
  --replace REPLACE     Template for complete string to replace (default:
                        {new_version})
  --current-version VERSION
                        Version that needs to be updated (default: None)
  --dry-run             Don't write any files, just pretend. (default: False)
  --new-version VERSION
                        New version that should be in the files (default:
                        None)
  --commit              Commit to version control (default: False)
  --no-commit           Do not commit to version control
  --tag                 Create a tag in version control (default: False)
  --no-tag              Do not create a tag in version control
  --sign-tags           Sign tags if created (default: False)
  --no-sign-tags        Do not sign tags if created
  --tag-name TAG_NAME   Tag name (only works with --tag) (default:
                        r{new_version})
  --tag-message TAG_MESSAGE
                        Tag message (default: [OPS] bumped version:
                        {current_version} → {new_version})
  --message COMMIT_MSG  Commit message (default: [OPS] bumped version:
                        {current_version} → {new_version})
""" % DESCRIPTION).lstrip()


def test_usage_string(tmpdir, capsys):
    tmpdir.chdir()

    with pytest.raises(SystemExit):
        main(["--help"])

    out, err = capsys.readouterr()
    assert err == ""

    for option_line in EXPECTED_OPTIONS:
        assert option_line in out, f"Usage string is missing {option_line}"

    assert EXPECTED_USAGE in out


def test_missing_explicit_config_file(tmpdir):
    tmpdir.chdir()
    with pytest.raises(argparse.ArgumentTypeError):
        main(["--config-file", "missing.cfg"])


def test_simple_replacement(tmpdir):
    tmpdir.join("VERSION").write("1.2.0")
    tmpdir.chdir()
    main(shlex_split("patch --current-version 1.2.0 --new-version 1.2.1 VERSION"))
    assert "1.2.1" == tmpdir.join("VERSION").read()


def test_simple_replacement_in_utf8_file(tmpdir):
    tmpdir.join("VERSION").write("Kröt1.3.0".encode(), "wb")
    tmpdir.chdir()
    tmpdir.join("VERSION").read("rb")
    main(shlex_split("patch --verbose --current-version 1.3.0 --new-version 1.3.1 VERSION"))
    out = tmpdir.join("VERSION").read("rb")
    assert "'Kr\\xc3\\xb6t1.3.1'" in repr(out)


def test_bump_version(tmpdir):
    tmpdir.join("file5").write("1.0.0")
    tmpdir.chdir()
    main(["patch", "--current-version", "1.0.0", "file5"])

    assert "1.0.1" == tmpdir.join("file5").read()


def test_bump_major(tmpdir):
    tmpdir.join("fileMAJORBUMP").write("4.2.8")
    tmpdir.chdir()
    main(["--current-version", "4.2.8", "major", "fileMAJORBUMP"])

    assert "5.0.0" == tmpdir.join("fileMAJORBUMP").read()


def test_non_existing_file(tmpdir):
    tmpdir.chdir()
    with pytest.raises(IOError):
        main(shlex_split("patch --current-version 1.2.0 --new-version 1.2.1 does_not_exist.txt"))


def test_non_existing_second_file(tmpdir):
    tmpdir.chdir()
    tmpdir.join("my_source_code.txt").write("1.2.3")
    with pytest.raises(IOError):
        main(shlex_split("patch --current-version 1.2.3 my_source_code.txt does_not_exist2.txt"))

    # first file is unchanged because second didn't exist
    assert "1.2.3" == tmpdir.join("my_source_code.txt").read()


def test_non_vcs_operations_if_vcs_is_not_installed(tmpdir, monkeypatch):
    monkeypatch.setenv("PATH", "")

    tmpdir.chdir()
    tmpdir.join("VERSION").write("31.0.3")

    main(["major", "--current-version", "31.0.3", "VERSION"])

    assert "32.0.0" == tmpdir.join("VERSION").read()


def test_log_no_config_file_info_message(tmpdir):
    tmpdir.chdir()

    tmpdir.join("a_file.txt").write("1.0.0")

    with LogCapture(level=logging.INFO) as log_capture:
        main(["--verbose", "--verbose", "--current-version", "1.0.0", "patch", 'a_file.txt'])

    log_capture.check_present(
        ("bumpsemver.cli", "INFO", "Could not read config file at .bumpsemver.cfg"),
        ("bumpsemver.version_part", "INFO", "Parsing version '1.0.0' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=1, minor=0, patch=0"),
        ("bumpsemver.cli", "INFO", "Attempting to increment part 'patch'"),
        ("bumpsemver.cli", "INFO", "Values are now: major=1, minor=0, patch=1"),
        ("bumpsemver.version_part", "INFO", "Parsing version '1.0.1' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=1, minor=0, patch=1"),
        ("bumpsemver.cli", "INFO", "New version will be '1.0.1'"),
        ("bumpsemver.cli", "INFO", "Asserting files a_file.txt contain the version string..."),
        ("bumpsemver.files.generic", "INFO", "Found '1.0.0' in a_file.txt at line 0: 1.0.0"),
        ("bumpsemver.files.generic", "INFO", "Changing generic file a_file.txt:"),
        ("bumpsemver.files.generic", "INFO", "--- a/a_file.txt\n+++ b/a_file.txt\n@@ -1 +1 @@\n-1.0.0\n+1.0.1"),
        ("bumpsemver.cli", "INFO", "Would write to config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO", "[bumpsemver]\ncurrent_version = 1.0.1\n\n"),
        order_matters=True
    )


def test_log_parse_doesnt_parse_current_version(tmpdir):
    tmpdir.chdir()

    with LogCapture() as log_capture:
        main(["--verbose", "--current-version", '12', "--new-version", "13", "patch"])

    log_capture.check_present(
        ("bumpsemver.cli", "INFO", "Could not read config file at .bumpsemver.cfg"),
        ("bumpsemver.version_part", "INFO", "Parsing version '12' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", 'WARNING',
         "Evaluating 'parse' option: '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)' does not parse current version '12'"),
        ("bumpsemver.version_part", "INFO", "Parsing version '13' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", 'WARNING',
         "Evaluating 'parse' option: '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)' does not parse current version '13'"),
        ("bumpsemver.cli", "INFO", "New version will be '13'"),
        ("bumpsemver.cli", "INFO", "Asserting files  contain the version string..."),
        ("bumpsemver.cli", "INFO", "Would write to config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO", "[bumpsemver]\ncurrent_version = 13\n\n"),
    )


def test_complex_info_logging(tmpdir):
    tmpdir.join("fileE").write("0.4.0")
    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write(dedent(r"""
        [bumpsemver]
        current_version = 0.4.0
        
        [bumpsemver:file:fileE]
        """).strip())

    with LogCapture() as log_capture:
        main(["patch", "--verbose"])

    log_capture.check(
        ("bumpsemver.cli", "INFO", "Reading config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO", "[bumpsemver]\ncurrent_version = 0.4.0\n\n[bumpsemver:file:fileE]"),
        ("bumpsemver.version_part", "INFO", "Parsing version '0.4.0' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=0, minor=4, patch=0"),
        ("bumpsemver.cli", "INFO", "Attempting to increment part 'patch'"),
        ("bumpsemver.cli", "INFO", "Values are now: major=0, minor=4, patch=1"),
        ("bumpsemver.version_part", "INFO", "Parsing version '0.4.1' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=0, minor=4, patch=1"),
        ("bumpsemver.cli", "INFO", "New version will be '0.4.1'"),
        ("bumpsemver.cli", "INFO", "Asserting files fileE contain the version string..."),
        ("bumpsemver.files.generic", "INFO", "Found '0.4.0' in fileE at line 0: 0.4.0"),
        ("bumpsemver.files.generic", "INFO", "Changing generic file fileE:"),
        ("bumpsemver.files.generic", "INFO", "--- a/fileE\n+++ b/fileE\n@@ -1 +1 @@\n-0.4.0\n+0.4.1"),
        ("bumpsemver.cli", "INFO", "Writing to config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO", "[bumpsemver]\ncurrent_version = 0.4.1\n\n[bumpsemver:file:fileE]\n\n")
    )


def test_multi_file_configuration(tmpdir):
    tmpdir.join("FULL_VERSION.txt").write("1.0.3")
    tmpdir.join("MAJOR_VERSION.txt").write("1.0.3")

    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 1.0.3

        [bumpsemver:file:FULL_VERSION.txt]

        [bumpsemver:file:MAJOR_VERSION.txt]
        """).strip())

    main(["major", "--verbose"])
    assert "2.0.0" in tmpdir.join("FULL_VERSION.txt").read()
    assert "2.0.0" in tmpdir.join("MAJOR_VERSION.txt").read()

    main(["patch"])
    assert "2.0.1" in tmpdir.join("FULL_VERSION.txt").read()
    assert "2.0.1" in tmpdir.join("MAJOR_VERSION.txt").read()


def test_search_replace_to_avoid_updating_unconcerned_lines(tmpdir):
    tmpdir.chdir()

    tmpdir.join("requirements.txt").write("Django>=1.5.6,<1.6\nMyProject==1.5.6")

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
      [bumpsemver]
      current_version = 1.5.6

      [bumpsemver:file:requirements.txt]
      search = MyProject=={current_version}
      replace = MyProject=={new_version}
      """).strip())

    with LogCapture() as log_capture:
        main(["minor", "--verbose"])

    log_capture.check(
        ("bumpsemver.cli", "INFO", "Reading config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO",
         "[bumpsemver]\ncurrent_version = 1.5.6\n\n[bumpsemver:file:requirements.txt]\n"
         "search = MyProject=={current_version}\nreplace = MyProject=={new_version}"),
        ("bumpsemver.version_part", "INFO", "Parsing version '1.5.6' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=1, minor=5, patch=6"),
        ("bumpsemver.cli", "INFO", "Attempting to increment part 'minor'"),
        ("bumpsemver.cli", "INFO", "Values are now: major=1, minor=6, patch=0"),
        ("bumpsemver.version_part", "INFO", "Parsing version '1.6.0' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=1, minor=6, patch=0"),
        ("bumpsemver.cli", "INFO", "New version will be '1.6.0'"),
        ("bumpsemver.cli", "INFO", "Asserting files requirements.txt contain the version string..."),
        ("bumpsemver.files.generic", "INFO", "Found 'MyProject==1.5.6' in requirements.txt at line 1: MyProject==1.5.6"),
        ("bumpsemver.files.generic", "INFO", "Changing generic file requirements.txt:"),
        ("bumpsemver.files.generic", "INFO",
         "--- a/requirements.txt\n+++ b/requirements.txt\n@@ -1,2 +1,2 @@\n Django>=1.5.6,<1.6\n-MyProject==1.5.6\n+MyProject==1.6.0"),
        ("bumpsemver.cli", "INFO", "Writing to config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO",
         "[bumpsemver]\ncurrent_version = 1.6.0\n\n[bumpsemver:file:requirements.txt]\n"
         "search = MyProject=={current_version}\nreplace = MyProject=={new_version}\n\n")
    )

    assert "MyProject==1.6.0" in tmpdir.join("requirements.txt").read()
    assert "Django>=1.5.6" in tmpdir.join("requirements.txt").read()


def test_search_replace_expanding_changelog(tmpdir):
    tmpdir.chdir()

    tmpdir.join("CHANGELOG.md").write(dedent("""
        My awesome software project Changelog
        =====================================
        
        Unreleased
        ----------
        
        * Some nice feature
        * Some other nice feature
        
        Version v8.1.1 (2014-05-28)
        ---------------------------
        
        * Another old nice feature
        
        """).strip())

    config_content = dedent("""
        [bumpsemver]
        current_version = 8.1.1

        [bumpsemver:file:CHANGELOG.md]
        search =
          Unreleased
          ----------
        replace =
          Unreleased
          ----------
          Version v{new_version} ({now:%Y-%m-%d})
          ---------------------------
        """).strip()

    tmpdir.join(".bumpsemver.cfg").write(config_content)

    with mock.patch("bumpsemver.cli.logger"):
        main(["minor", "--verbose"])

    predate = dedent("""
        Unreleased
        ----------
        Version v8.2.0 (20
        """).strip()

    postdate = dedent("""
        )
        ---------------------------

        * Some nice feature
        * Some other nice feature
        """).strip()

    assert predate in tmpdir.join("CHANGELOG.md").read()
    assert postdate in tmpdir.join("CHANGELOG.md").read()


def test_non_matching_search_does_not_modify_file(tmpdir):
    tmpdir.chdir()

    changelog_content = dedent("""
        # Unreleased
        
        * bullet point A
        
        # Release v'older' (2019-09-17)
        
        * bullet point B
        """).strip()

    config_content = dedent("""
        [bumpsemver]
        current_version = 1.0.3
        
        [bumpsemver:file:CHANGELOG.md]
        search = Not-yet-released
        replace = Release v{new_version} ({now:%Y-%m-%d})
        """).strip()

    tmpdir.join("CHANGELOG.md").write(changelog_content)
    tmpdir.join(".bumpsemver.cfg").write(config_content)

    with pytest.raises(exceptions.VersionNotFoundException, match="Did not find 'Not-yet-released' in file: 'CHANGELOG.md'"):
        main(["patch", "--verbose"])

    assert changelog_content == tmpdir.join("CHANGELOG.md").read()
    assert config_content in tmpdir.join(".bumpsemver.cfg").read()


def test_search_replace_cli(tmpdir):
    tmpdir.join("file89").write("My birthday: 3.5.98\nCurrent version: 3.5.98")
    tmpdir.chdir()
    main([
        "--current-version", '3.5.98',
        "--search", "Current version: {current_version}",
        "--replace", "Current version: {new_version}",
        "minor",
        "file89",
    ])

    assert "My birthday: 3.5.98\nCurrent version: 3.6.0" == tmpdir.join("file89").read()


def test_deprecation_warning_files_in_global_configuration(tmpdir):
    tmpdir.chdir()

    tmpdir.join("fileX").write("3.2.1")
    tmpdir.join("fileY").write("3.2.1")
    tmpdir.join("fileZ").write("3.2.1")

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 3.2.1
        files = fileX fileY fileZ
        """).strip())

    warning_registry = getattr(bumpsemver, "__warningregistry__", None)
    if warning_registry:
        warning_registry.clear()
    warnings.resetwarnings()
    warnings.simplefilter("always")
    with warnings.catch_warnings(record=True) as received_warnings:
        main(["patch"])

    w = received_warnings.pop()
    assert issubclass(w.category, PendingDeprecationWarning)
    assert "'files =' configuration will be deprecated, please use" in str(w.message)


def test_deprecation_warning_multiple_files_cli(tmpdir):
    tmpdir.chdir()

    tmpdir.join("fileA").write("1.2.3")
    tmpdir.join("fileB").write("1.2.3")
    tmpdir.join("fileC").write("1.2.3")

    warning_registry = getattr(bumpsemver, "__warningregistry__", None)
    if warning_registry:
        warning_registry.clear()
    warnings.resetwarnings()
    warnings.simplefilter("always")
    with warnings.catch_warnings(record=True) as received_warnings:
        main(["--current-version", "1.2.3", "patch", "fileA", "fileB", "fileC"])

    w = received_warnings.pop()
    assert issubclass(w.category, PendingDeprecationWarning)
    assert "Giving multiple files on the command line will be deprecated" in str(w.message)


def test_multi_line_search_is_found(tmpdir):
    tmpdir.chdir()

    tmpdir.join("the_alphabet.txt").write(dedent("""
        A
        B
        C
        """))

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 9.8.7
        
        [bumpsemver:file:the_alphabet.txt]
        search =
          A
          B
          C
        replace =
          A
          B
          C
          {new_version}
        """).strip())

    main(['major'])

    assert dedent("""
      A
      B
      C
      10.0.0
    """) == tmpdir.join("the_alphabet.txt").read()


def test_configparser_empty_lines_in_values(tmpdir):
    tmpdir.chdir()

    tmpdir.join("CHANGES.rst").write(dedent("""
    My changelog
    ============

    current
    -------

    """))

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
    [bumpsemver]
    current_version = 0.4.1

    [bumpsemver:file:CHANGES.rst]
    search =
      current
      -------
    replace = current
      -------


      {new_version}
      -------
      """).strip())

    main(["patch"])
    assert dedent("""
      My changelog
      ============
      current
      -------


      0.4.2
      -------

    """) == tmpdir.join("CHANGES.rst").read()


def test_regression_dont_touch_capitalization_of_keys_in_config(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 0.1.0
    
        [other]
        DJANGO_SETTINGS = Value
        """).strip())

    main(["patch"])

    assert dedent("""
        [bumpsemver]
        current_version = 0.1.1
    
        [other]
        DJANGO_SETTINGS = Value
        """).strip() == tmpdir.join(".bumpsemver.cfg").read().strip()


def test_regression_new_version_cli_in_files(tmpdir):
    tmpdir.chdir()
    tmpdir.join("myp___init__.py").write("__version__ = '0.7.2'")
    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 0.7.2
        message = v{new_version}
        tag_name = {new_version}
        tag = true
        commit = true
        [bumpsemver:file:myp___init__.py]
        """).strip())

    main("patch --allow-dirty --verbose --new-version 0.9.3".split(" "))

    assert "__version__ = '0.9.3'" == tmpdir.join("myp___init__.py").read()
    assert "current_version = 0.9.3" in tmpdir.join(".bumpsemver.cfg").read()


def test_correct_interpolation_for_setup_cfg_files(tmpdir):
    tmpdir.chdir()
    tmpdir.join("file.py").write("XX-XX-XXXX v. X.X.X")
    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 0.7.2
        search = XX-XX-XXXX v. X.X.X
        replace = {now:%m-%d-%Y} v. {new_version}
        [bumpsemver:file:file.py]
        """).strip())

    main(["major"])

    assert datetime.now().strftime("%m-%d-%Y") + " v. 1.0.0" == tmpdir.join("file.py").read()
    assert "current_version = 1.0.0" in tmpdir.join(".bumpsemver.cfg").read()


class TestSplitArgsInOptionalAndPositional:

    def test_all_optional(self):
        params = ["--allow-dirty", "--verbose", "--dry-run", "--tag-name", '"Tag"']
        positional, optional = split_args_in_optional_and_positional(params)

        assert positional == []
        assert optional == params

    def test_all_positional(self):
        params = ["minor", "setup.py"]
        positional, optional = split_args_in_optional_and_positional(params)

        assert positional == params
        assert optional == []

    def test_no_args(self):
        assert split_args_in_optional_and_positional([]) == ([], [])

    def test_1optional_2positional(self):
        params = ["--dry-run", "major", "setup.py"]
        positional, optional = split_args_in_optional_and_positional(params)

        assert positional == ['major', 'setup.py']
        assert optional == ["--dry-run"]

    def test_2optional_1positional(self):
        params = ["--dry-run", "--message", '"Commit"', "major"]
        positional, optional = split_args_in_optional_and_positional(params)

        assert positional == ['major']
        assert optional == ["--dry-run", "--message", '"Commit"']

    def test_2optional_mixed_2positional(self):
        params = ["--allow-dirty", "--message", '"Commit"', "minor", "setup.py"]
        positional, optional = split_args_in_optional_and_positional(params)

        assert positional == ["minor", "setup.py"]
        assert optional == ["--allow-dirty", "--message", '"Commit"']


def test_defaults_in_usage_with_config(tmpdir, capsys):
    tmpdir.chdir()
    tmpdir.join("my_defaults.cfg").write(dedent("""
        [bumpsemver]
        current_version: 18
        new_version: 19
        [bumpsemver:file:file1]
        [bumpsemver:file:file2]
        [bumpsemver:file:file3]
        """).strip())
    with pytest.raises(SystemExit):
        main(["--config-file", "my_defaults.cfg", "--help"])

    out, err = capsys.readouterr()

    assert "Version that needs to be updated (default: 18)" in out
    assert "New version that should be in the files (default: 19)" in out
    assert "[--current-version VERSION]" in out
    assert "[--new-version VERSION]" in out
    assert "[file ...]" in out


def test_config_file(tmpdir):
    tmpdir.join("file1").write("0.9.34")
    tmpdir.join("my_bump_config.cfg").write(dedent("""
        [bumpsemver]
        current_version: 0.9.34
        new_version: 0.9.35
        [bumpsemver:file:file1]
        """).strip())

    tmpdir.chdir()
    main(shlex_split("patch --config-file my_bump_config.cfg"))

    assert "0.9.35" == tmpdir.join("file1").read()


def test_default_config_files(tmpdir):
    tmpdir.join("file2").write("0.10.2")
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version: 0.10.2
        new_version: 0.10.3
        [bumpsemver:file:file2]
        """).strip())

    tmpdir.chdir()
    main(["patch"])

    assert "0.10.3" == tmpdir.join("file2").read()


def test_file_keyword_with_suffix_is_accepted(tmpdir, file_keyword):
    tmpdir.join("file2").write("0.10.2")
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version: 0.10.2
        new_version: 0.10.3
        [bumpsemver:%s:file2]
        """ % file_keyword).strip())

    tmpdir.chdir()
    main(["patch"])

    assert "0.10.3" == tmpdir.join("file2").read()


def test_config_file_is_updated(tmpdir):
    tmpdir.join("file3").write("0.0.13")
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version: 0.0.13
        new_version: 0.0.14
        [bumpsemver:file:file3]
        """).strip())

    tmpdir.chdir()
    main(["patch", "--verbose"])

    assert """[bumpsemver]
current_version = 0.0.14

[bumpsemver:file:file3]
""" == tmpdir.join(".bumpsemver.cfg").read()


@pytest.mark.parametrize("newline", [b"\n", b"\r\n"])
def test_retain_newline(tmpdir, newline):
    tmpdir.join("file.py").write_binary(dedent("""
        0.7.2
        Some Content
        """).strip().encode(encoding='UTF-8').replace(b"\n", newline))
    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write_binary(dedent("""
        [bumpsemver]
        current_version = 0.7.2
        search = {current_version}
        replace = {new_version}
        [bumpsemver:file:file.py]
        """).strip().encode(encoding='UTF-8').replace(b"\n", newline))

    main(["major"])

    assert newline in tmpdir.join("file.py").read_binary()
    new_config = tmpdir.join(".bumpsemver.cfg").read_binary()
    assert newline in new_config

    # Ensure there is only a single newline (not two) at the end of the file and that it is of the right type
    assert new_config.endswith(b"[bumpsemver:file:file.py]" + newline)



import io
from abc import ABCMeta, abstractmethod
from difflib import unified_diff

from bumpsemver.version_part import Version


class FileTypeBase:
    __metaclass__ = ABCMeta

    def __init__(self, path, version_config, file_type, logger):
        self.path = path
        self._version_config = version_config
        self.file_type = file_type
        self.logger = logger

    @abstractmethod
    def should_contain_version(self, version: Version, context: dict) -> None:
        """
        Raise VersionNotFound if the parsed version isn't present in this file.
        Otherwise, return if the version number is present.
        """

    @abstractmethod
    def contains(self, search: str) -> bool:
        """
        Return True if the version string is present, otherwise, False
        """

    @abstractmethod
    def replace(self, current_version: Version, new_version: Version, context: dict, dry_run: bool) -> None:
        """
        Update the version if it is not a dry run
        """

    def update_file(self, file_content_before: str, file_content_after: str, dry_run: bool) -> None:
        """
        Write changes to the file if it is not a dry run
        """
        with open(self.path, "rt", encoding="utf-8") as orig_fp:
            _dummy = orig_fp.read()
            file_new_lines = orig_fp.newlines

        if file_content_before != file_content_after:
            # reassemble the file to retain the original os-specific newline separator
            self.logger.info(f"{'Would change' if dry_run else 'Changing'} {self.file_type} file {self.path}:")
            self.logger.info("\n".join(list(unified_diff(
                file_content_before.splitlines(),
                file_content_after.splitlines(),
                lineterm="",
                fromfile=f"a/{self.path}",
                tofile=f"b/{self.path}"
            ))))
        else:
            self.logger.info(f"{'Would not change' if dry_run else 'Not changing'} {self.file_type} file {self.path}")

        if not dry_run:
            with io.open(self.path, "wt", encoding="utf-8", newline=file_new_lines) as orig_fp:
                orig_fp.write(file_content_after)

    def __str__(self):
        return self.path

    @abstractmethod
    def __repr__(self):
        pass



import argparse
import io
import logging
import os
import re
import sre_constants
import sys
import warnings
from configparser import (RawConfigParser, NoOptionError)
from datetime import datetime

from bumpsemver import __version__, __title__
from bumpsemver.exceptions import (
    IncompleteVersionRepresentationException,
    MissingValueForSerializationException,
    WorkingDirectoryIsDirtyException,
)
from bumpsemver.files.generic import ConfiguredGenericFile
from bumpsemver.files.json import ConfiguredJSONFile
from bumpsemver.files.yaml import ConfiguredYAMLFile
from bumpsemver.git import Git
from bumpsemver.utils import key_value_string
from bumpsemver.version_part import VersionConfig

python_version = sys.version.split("\n")[0].split(" ")[0]
DESCRIPTION = f"{__title__}: v{__version__} (using Python v{python_version})"


# detect either:
# bumpsemver:yaml:value
# bumpsemver:yaml(suffix):value
# bumpsemver:yaml ( suffix with spaces):value
# bumpsemver:json:value
# bumpsemver:json(suffix):value
# bumpsemver:json ( suffix with spaces):value
# bumpsemver:file:value
# bumpsemver:file(suffix):value
# bumpsemver:file ( suffix with spaces):value
RE_DETECT_SECTION_TYPE = re.compile(
    r"^bumpsemver:("
    r"(?P<file>file)(\s*\(\s*(?P<file_suffix>[^):]+)\)?)?"
    r"|"
    r"(?P<json>json)(\s*\(\s*(?P<json_suffix>[^):]+)\)?)?"
    r"|"
    r"(?P<yaml>yaml)(\s*\(\s*(?P<yaml_suffix>[^):]+)\)?)?"
    r"):(?P<value>.+)",
)

logger = logging.getLogger(__name__)
time_context = {"now": datetime.now(), "utcnow": datetime.utcnow()}

OPTIONAL_ARGUMENTS_THAT_TAKE_VALUES = [
    "--config-file",
    "--current-version",
    "--message",
    "--new-version",
    "--search",
    "--replace",
    "--tag-name",
    "--tag-message",
]


def main(original_args=None):
    #
    # determine configuration based on command-line arguments and on-disk configuration files
    args, known_args, root_parser, positionals = _parse_arguments_phase_1(original_args)
    _setup_logging(known_args.verbose)
    vcs_info = _determine_vcs_usability()
    defaults = _determine_current_version(vcs_info)
    explicit_config = None
    if hasattr(known_args, "config_file"):
        explicit_config = known_args.config_file
    config_file = _determine_config_file(explicit_config)
    config, config_file_exists, config_newlines, part_configs, files = _load_configuration(config_file, explicit_config, defaults)
    known_args, parser2, remaining_argv = _parse_arguments_phase_2(args, defaults, root_parser)
    version_config = _setup_version_config(known_args, part_configs)
    current_version = version_config.parse(known_args.current_version)
    context = {**time_context, **vcs_info}
    #
    # calculate the desired new version
    new_version = _assemble_new_version(context, current_version, defaults, known_args.current_version, positionals, version_config)
    args, file_names = _parse_arguments_phase_3(remaining_argv, positionals, defaults, parser2)
    new_version = _parse_new_version(args, new_version, version_config)

    # replace version in target files
    vcs = _determine_vcs_dirty(defaults)
    files.extend(ConfiguredGenericFile(file_name, version_config) for file_name in (file_names or positionals[1:]))
    _check_files_contain_version(files, current_version, context)
    _replace_version_in_files(files, current_version, new_version, args.dry_run, context)
    config.remove_option("bumpsemver", "new_version")

    # store the new version
    _update_config_file(config, config_file, config_newlines, config_file_exists, args.new_version, args.dry_run)

    # commit and tag
    if vcs:
        context = _commit_to_vcs(files, config_file, config_file_exists, vcs, args, current_version, new_version)
        _tag_in_vcs(vcs, context, args)


def split_args_in_optional_and_positional(args):
    # manually parsing positional arguments because stupid argparse can't mix positional and optional arguments

    positions = []
    for i, arg in enumerate(args):

        previous = None

        if i > 0:
            previous = args[i - 1]

        if (not arg.startswith("-")) and (previous not in OPTIONAL_ARGUMENTS_THAT_TAKE_VALUES):
            positions.append(i)

    positionals = [arg for i, arg in enumerate(args) if i in positions]
    args = [arg for i, arg in enumerate(args) if i not in positions]

    return positionals, args


def _parse_arguments_phase_1(original_args):
    positionals, args = split_args_in_optional_and_positional(sys.argv[1:] if original_args is None else original_args)
    if len(positionals[1:]) > 2:
        warnings.warn(
            "Giving multiple files on the command line will be deprecated, please use [bumpsemver:file:...] in a config file.",
            PendingDeprecationWarning,
        )
    root_parser = argparse.ArgumentParser(add_help=False)
    root_parser.add_argument(
        "--config-file",
        metavar="FILE",
        default=argparse.SUPPRESS,
        required=False,
        help="Config file to read most of the variables from (default: .bumpsemver.cfg)",
    )
    root_parser.add_argument(
        "--verbose",
        action="count",
        default=0,
        help="Print verbose logging to stderr",
        required=False,
    )
    root_parser.add_argument(
        "--allow-dirty",
        action="store_true",
        default=False,
        help="Don't abort if working directory is dirty",
        required=False,
    )
    known_args, _ = root_parser.parse_known_args(args)
    return args, known_args, root_parser, positionals


def _setup_logging(verbose):
    try:
        log_level = [logging.WARNING, logging.INFO, logging.DEBUG][verbose]
    except IndexError:
        log_level = logging.DEBUG
    root_logger = logging.getLogger("")
    root_logger.setLevel(log_level)
    logger.debug("Starting {}‚", DESCRIPTION)


def _determine_vcs_usability():
    vcs_info = {}
    if Git.is_usable():
        vcs_info.update(Git.latest_tag_info())
    return vcs_info


def _determine_current_version(vcs_info):
    defaults = {}
    if "current_version" in vcs_info:
        defaults["current_version"] = vcs_info["current_version"]
    return defaults


def _determine_config_file(explicit_config):
    if explicit_config:
        return explicit_config
    return ".bumpsemver.cfg"


def _load_configuration(config_file, explicit_config, defaults):
    # noinspection PyTypeChecker
    config = RawConfigParser("")
    # don't transform keys to lowercase (which would be the default)
    config.optionxform = lambda option: option
    config.add_section("bumpsemver")
    config_file_exists = os.path.exists(config_file)

    if not config_file_exists:
        message = f"Could not read config file at {config_file}"
        if explicit_config:
            raise argparse.ArgumentTypeError(message)
        logger.info(message)
        return config, config_file_exists, None, {}, []

    logger.info(f"Reading config file {config_file}:")

    with open(config_file, "rt", encoding="utf-8") as config_fp:
        config_content = config_fp.read()
        config_newlines = config_fp.newlines

    logger.info(config_content)
    config.read_string(config_content)
    log_config = io.StringIO()
    config.write(log_config)

    if config.has_option("bumpsemver", "files"):
        warnings.warn(
            "'files =' configuration will be deprecated, please use [bumpsemver:file:...]",
            PendingDeprecationWarning,
        )

    defaults.update(dict(config.items("bumpsemver")))

    for bool_value_name in ("commit", "tag", "dry_run"):
        try:
            defaults[bool_value_name] = config.getboolean("bumpsemver", bool_value_name)
        except NoOptionError:
            pass  # no default value then

    part_configs = {}
    files = []

    for section_name in config.sections():
        section_type_match = RE_DETECT_SECTION_TYPE.match(section_name)

        if not section_type_match:
            continue

        section_type = section_type_match.groupdict()
        section_value = section_type.get("value")
        section_config = dict(config.items(section_name))

        filename = section_value

        section_config["part_configs"] = part_configs

        if section_type.get("file"):
            if "search" not in section_config:
                section_config["search"] = defaults.get("search", "{current_version}")
            if "replace" not in section_config:
                section_config["replace"] = defaults.get("replace", "{new_version}")
            files.append(ConfiguredGenericFile(filename, VersionConfig(**section_config)))
        elif section_type.get("json"):
            jsonpath = section_config.pop("jsonpath", defaults.get("jsonpath", "version"))
            files.append(ConfiguredJSONFile(filename, jsonpath, VersionConfig(**section_config)))
        elif section_type.get("yaml"):
            yamlpath = section_config.pop("yamlpath", defaults.get("yamlpath", "version"))
            files.append(ConfiguredYAMLFile(filename, yamlpath, VersionConfig(**section_config)))
        else:
            # the other cases must not be possible, because the regex matching at the beginning of this function should filter them out
            pass

    return config, config_file_exists, config_newlines, part_configs, files


def _parse_arguments_phase_2(args, defaults, root_parser):
    parser2 = argparse.ArgumentParser(prog="bumpsemver", add_help=False, parents=[root_parser])
    parser2.set_defaults(**defaults)
    parser2.add_argument(
        "--current-version",
        metavar="VERSION",
        help="Version that needs to be updated",
        required=False,
    )
    parser2.add_argument(
        "--search",
        metavar="SEARCH",
        help="Template for complete string to search",
        default=defaults.get("search", "{current_version}"),
    )
    parser2.add_argument(
        "--replace",
        metavar="REPLACE",
        help="Template for complete string to replace",
        default=defaults.get("replace", "{new_version}"),
    )
    known_args, remaining_argv = parser2.parse_known_args(args)

    defaults.update(vars(known_args))

    return known_args, parser2, remaining_argv


def _setup_version_config(known_args, part_configs):
    try:
        version_config = VersionConfig(search=known_args.search, replace=known_args.replace, part_configs=part_configs)
    except sre_constants.error:
        sys.exit(1)
    return version_config


def _assemble_new_version(context, current_version, defaults, arg_current_version, positionals, version_config):
    new_version = None
    if "new_version" not in defaults and arg_current_version:
        try:
            if current_version and positionals:
                logger.info(f"Attempting to increment part '{positionals[0]}'")
                new_version = current_version.bump(positionals[0], version_config.order())
                logger.info(f"Values are now: {key_value_string(new_version.values)}")
                defaults["new_version"] = version_config.serialize(new_version, context)
        except MissingValueForSerializationException as err:
            logger.info(f"Opportunistic finding of new_version failed: {err.message}")
        except IncompleteVersionRepresentationException as err:
            logger.info(f"Opportunistic finding of new_version failed: {err.message}")
        except KeyError:
            logger.info("Opportunistic finding of new_version failed")
    return new_version


def _parse_arguments_phase_3(remaining_argv, positionals, defaults, parser2):
    parser3 = argparse.ArgumentParser(
        prog="bumpsemver",
        description=DESCRIPTION,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        conflict_handler="resolve",
        parents=[parser2],
    )
    parser3.set_defaults(**defaults)
    parser3.add_argument(
        "--current-version",
        metavar="VERSION",
        help="Version that needs to be updated",
        required="current_version" not in defaults,
    )
    parser3.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="Don't write any files, just pretend.",
    )
    parser3.add_argument(
        "--new-version",
        metavar="VERSION",
        help="New version that should be in the files",
        required="new_version" not in defaults,
    )
    commit_group = parser3.add_mutually_exclusive_group()
    commit_group.add_argument(
        "--commit",
        action="store_true",
        dest="commit",
        help="Commit to version control",
        default=defaults.get("commit", False),
    )
    commit_group.add_argument(
        "--no-commit",
        action="store_false",
        dest="commit",
        help="Do not commit to version control",
        default=argparse.SUPPRESS,
    )
    tag_group = parser3.add_mutually_exclusive_group()
    tag_group.add_argument(
        "--tag",
        action="store_true",
        dest="tag",
        default=defaults.get("tag", False),
        help="Create a tag in version control",
    )
    tag_group.add_argument(
        "--no-tag",
        action="store_false",
        dest="tag",
        help="Do not create a tag in version control",
        default=argparse.SUPPRESS,
    )
    sign_tags_group = parser3.add_mutually_exclusive_group()
    sign_tags_group.add_argument(
        "--sign-tags",
        action="store_true",
        dest="sign_tags",
        help="Sign tags if created",
        default=defaults.get("sign_tags", False),
    )
    sign_tags_group.add_argument(
        "--no-sign-tags",
        action="store_false",
        dest="sign_tags",
        help="Do not sign tags if created",
        default=argparse.SUPPRESS,
    )
    parser3.add_argument(
        "--tag-name",
        metavar="TAG_NAME",
        help="Tag name (only works with --tag)",
        default=defaults.get("tag_name", "r{new_version}"),
    )
    parser3.add_argument(
        "--tag-message",
        metavar="TAG_MESSAGE",
        dest="tag_message",
        help="Tag message",
        default=defaults.get(
            "tag_message",
            "[OPS] bumped version: {current_version} → {new_version}"
        ),
    )
    parser3.add_argument(
        "--message",
        metavar="COMMIT_MSG",
        help="Commit message",
        default=defaults.get(
            "message", "[OPS] bumped version: {current_version} → {new_version}"
        ),
    )
    file_names = []
    if "files" in defaults:
        assert defaults["files"] is not None
        file_names = defaults["files"].split(" ")
    parser3.add_argument("part", help="Part of the version to be bumped.")
    parser3.add_argument(
        "files", metavar="file", nargs="*", help="Files to change",
        default=file_names
    )
    args = parser3.parse_args(remaining_argv + positionals)

    if args.dry_run:
        logger.info("Dry run active, won't touch any files.")

    return args, file_names


def _parse_new_version(args, new_version, version_config):
    if args.new_version:
        new_version = version_config.parse(args.new_version)
    logger.info(f"New version will be '{args.new_version}'")
    return new_version


def _determine_vcs_dirty(defaults):
    if not Git.is_usable():
        return None

    try:
        Git.assert_non_dirty()
    except WorkingDirectoryIsDirtyException as err:
        if defaults["allow_dirty"]:
            return None
        logger.warning(f"{err.message}\n\nUse --allow-dirty to override this if you know what you're doing.")
        raise

    return Git


def _check_files_contain_version(files, current_version, context):
    # make sure files exist and contain version string
    logger.info(f"Asserting files {', '.join([str(f) for f in files])} contain the version string...")
    for file_item in files:
        file_item.should_contain_version(current_version, context)


def _replace_version_in_files(files, current_version, new_version, dry_run, context):
    # change version string in files
    for file_item in files:
        file_item.replace(current_version, new_version, context, dry_run)


def _update_config_file(config, config_file, config_newlines, config_file_exists, new_version, dry_run):
    config.set("bumpsemver", "current_version", new_version)
    new_config = io.StringIO()
    try:
        write_to_config_file = (not dry_run) and config_file_exists

        logger.info(f"{'Would write' if not write_to_config_file else 'Writing'} to config file {config_file}:")

        config.write(new_config)
        logger.info(new_config.getvalue())

        if write_to_config_file:
            with open(config_file, "wt", encoding="utf-8", newline=config_newlines) as config_fp:
                config_fp.write(new_config.getvalue().strip() + "\n")

    except UnicodeEncodeError:
        warnings.warn(
            "Unable to write UTF-8 to config file, because of an old configparser version. "
            "Update with `pip install --upgrade configparser`."
        )


def _commit_to_vcs(files, config_file, config_file_exists, vcs, args, current_version, new_version):
    commit_files = [f.path for f in files]
    if config_file_exists:
        commit_files.append(config_file)
    assert vcs.is_usable(), f"Did find '{vcs.__name__}' unusable, unable to commit."
    do_commit = args.commit and not args.dry_run
    logger.info(f"{'Would prepare' if not do_commit else 'Preparing'} {vcs.__name__} commit")
    for path in commit_files:
        logger.info(f"{'Would add' if not do_commit else 'Adding'} changes in file '{path}' to {vcs.__name__}")

        if do_commit:
            vcs.add_path(path)

    context = {
        "current_version": args.current_version,
        "new_version": args.new_version,
    }
    context.update(time_context)
    context.update({f"current_{part}": current_version[part].value for part in current_version})
    context.update({f"new_{part}": new_version[part].value for part in new_version})

    commit_message = args.message.format(**context)

    logger.info(f"{'Would commit' if not do_commit else 'Committing'} to {vcs.__name__} with message '{commit_message}'")
    if do_commit:
        vcs.commit(message=commit_message, context=context)
    return context


def _tag_in_vcs(vcs, context, args):
    sign_tags = args.sign_tags
    tag_name = args.tag_name.format(**context)
    tag_message = args.tag_message.format(**context)
    do_tag = args.tag and not args.dry_run
    logger.info(f"{'Would tag' if not do_tag else 'Tagging'} `{tag_name}` "
                f"{f'with message `{tag_message}`' if tag_message else 'without message'} "
                f"in {vcs.__name__} and {'signing' if sign_tags else 'not signing'}")
    if do_tag:
        vcs.tag(sign_tags, tag_name, tag_message)



# pylint: skip-file

# noinspection PyPackageRequirements
import pytest
# noinspection PyPackageRequirements
from testfixtures import LogCapture

import json
from textwrap import dedent

from bumpsemver import exceptions
from bumpsemver.cli import main


@pytest.fixture(params=[
    "json",
    "json(suffix)",
    "json (suffix with space)",
    "json (suffix lacking closing paren",
])
def json_keyword(request):
    """Return multiple possible styles for the bumpsemver:json keyword."""
    return request.param


def test_json_file_simple(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 75.0.1
        [bumpsemver:json:package.json]
        jsonpath = version
        """).strip())
    tmpdir.join("package.json").write("""
        {
          "version": "75.0.1"
        }
    """)
    main(["minor"])
    assert "current_version = 75.1.0" in tmpdir.join(".bumpsemver.cfg").read()
    data = json.loads(tmpdir.join("package.json").read())
    assert data["version"] == "75.1.0"


def test_json_file_exact_path(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 75.2.1
        [bumpsemver:json:package.json]
        jsonpath = version
        """).strip())
    tmpdir.join("package.json").write("""
        {
          "version": "75.2.1",
          "dependencies": {
            "@babel/code-frame": {
              "version": "75.2.1",
              "requires": {
                "version": "^75.2.1"
              }
            }
          }
        }
    """)
    main(["minor"])
    assert "current_version = 75.3.0" in tmpdir.join(".bumpsemver.cfg").read()
    data = json.loads(tmpdir.join("package.json").read())
    assert data['version'] == "75.3.0"
    assert data['dependencies']['@babel/code-frame']['version'] == "75.2.1"
    assert data['dependencies']['@babel/code-frame']['requires']['version'] == "^75.2.1"


def test_json_file_with_suffix_two(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 75.4.1
        [bumpsemver:json(a):package.json]
        jsonpath = version
        [bumpsemver:json(b):package.json]
        jsonpath = dependencies."@babel/code-frame".version
        """).strip())
    tmpdir.join("package.json").write("""
        {
          "version": "75.4.1",
          "dependencies": {
            "@babel/code-frame": {
              "version": "75.4.1",
              "requires": {
                "version": "75.4.1"
              }
            }
          }
        }
    """)
    main(["minor"])
    assert "current_version = 75.5.0" in tmpdir.join(".bumpsemver.cfg").read()
    data = json.loads(tmpdir.join("package.json").read())
    assert data['version'] == "75.5.0"
    assert data['dependencies']['@babel/code-frame']['version'] == "75.5.0"
    assert data['dependencies']['@babel/code-frame']['requires']['version'] == "75.4.1"


def test_json_file_with_suffix_complex(tmpdir, json_keyword):
    tmpdir.join("file2").write('{"version": "5.10.2"}')
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version: 5.10.2
        new_version: 5.10.8
        [bumpsemver:%s:file2]
        jsonpath: version
        """ % json_keyword).strip())
    tmpdir.chdir()
    main(["patch"])

    assert "5.10.8" in tmpdir.join("file2").read()


def test_json_file_incl_array(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 75.6.1
        [bumpsemver:json:package.json]
        jsonpath = dependencies[1].*.version
        """).strip())
    tmpdir.join("package.json").write("""
        {
          "version": "75.6.1",
          "dependencies": [
            {
              "@babel/code-frame": {
                "version": "75.6.1"
              }
            },
            {
              "@babel/highlight": {
                "version": "75.6.1"
              }
            }
          ]
        }
    """)
    main(["minor"])
    assert "current_version = 75.7.0" in tmpdir.join(".bumpsemver.cfg").read()
    data = json.loads(tmpdir.join("package.json").read())
    assert data['version'] == "75.6.1"
    assert data['dependencies'][0]['@babel/code-frame']['version'] == "75.6.1"
    assert data['dependencies'][1]['@babel/highlight']['version'] == "75.7.0"


def test_json_file_key_does_not_exist(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 75.8.1
        [bumpsemver:json:package.json]
        jsonpath = version_foo_bar
        """).strip())
    tmpdir.join("package.json").write("""
        {
          "version": "75.8.1"
        }
    """)

    with pytest.raises(exceptions.VersionNotFoundException,
                       match="Did not find '75.8.1' at jsonpath 'version_foo_bar' in file: 'package.json'"):
        main(["minor"])

    assert "current_version = 75.8.1" in tmpdir.join(".bumpsemver.cfg").read()
    data = json.loads(tmpdir.join("package.json").read())
    assert data['version'] == "75.8.1"


def test_json_file_key_exists_but_wrong_value(tmpdir):
    tmpdir.chdir()
    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 75.10.1
        [bumpsemver:json:package.json]
        jsonpath = version
        """).strip())
    tmpdir.join("package.json").write("""
        {
          "version": "^75.10.1"
        }
    """)

    with pytest.raises(exceptions.VersionNotFoundException, match="Did not find '75.10.1' at jsonpath 'version' in file: 'package.json'"):
        main(["minor"])

    assert "current_version = 75.10.1" in tmpdir.join(".bumpsemver.cfg").read()
    data = json.loads(tmpdir.join("package.json").read())
    assert data['version'] == "^75.10.1"


def test_json_file_info_logging(tmpdir):
    tmpdir.join("fileJ").write(dedent("""
        {
          "version": "0.4.0"
        }
        """).strip())
    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write(dedent(r"""
        [bumpsemver]
        current_version = 0.4.0

        [bumpsemver:json:fileJ]
        jsonpath = version
        """).strip())

    with LogCapture() as log_capture:
        main(["patch", "--verbose"])

    log_capture.check(
        ("bumpsemver.cli", "INFO", "Reading config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO", "[bumpsemver]\ncurrent_version = 0.4.0\n\n[bumpsemver:json:fileJ]\njsonpath = version"),
        ("bumpsemver.version_part", "INFO", "Parsing version '0.4.0' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=0, minor=4, patch=0"),
        ("bumpsemver.cli", "INFO", "Attempting to increment part 'patch'"),
        ("bumpsemver.cli", "INFO", "Values are now: major=0, minor=4, patch=1"),
        ("bumpsemver.version_part", "INFO", "Parsing version '0.4.1' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=0, minor=4, patch=1"),
        ("bumpsemver.cli", "INFO", "New version will be '0.4.1'"),
        ("bumpsemver.cli", "INFO", "Asserting files fileJ contain the version string..."),
        ("bumpsemver.files.json", "INFO", "Changing json file fileJ:"),
        ("bumpsemver.files.json", "INFO",
         "--- a/fileJ\n+++ b/fileJ\n@@ -1,3 +1,3 @@\n {\n-  \"version\": \"0.4.0\"\n+  \"version\": \"0.4.1\"\n }"),
        ("bumpsemver.cli", "INFO", "Writing to config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO", "[bumpsemver]\ncurrent_version = 0.4.1\n\n[bumpsemver:json:fileJ]\njsonpath = version\n\n")
    )



import errno
import logging
import os
import subprocess
from tempfile import NamedTemporaryFile

from bumpsemver.exceptions import (WorkingDirectoryIsDirtyException)

logger = logging.getLogger(__name__)


class Git:

    @classmethod
    def commit(cls, message, context, extra_args=None):
        extra_args = extra_args or []
        with NamedTemporaryFile("wb", delete=False) as temp_fp:
            temp_fp.write(message.encode("utf-8"))
        env = os.environ.copy()
        for key in ("current_version", "new_version"):
            env[str("BUMPSEMVER_" + key.upper())] = str(context[key])
        try:
            subprocess.check_output(["git", "commit", "-F", temp_fp.name] + extra_args, env=env)
        except subprocess.CalledProcessError as exc:
            err_msg = f"Failed to run {exc.cmd}: return code {exc.returncode}, output: {exc.output}"
            logger.exception(err_msg)
            raise exc
        finally:
            os.unlink(temp_fp.name)

    @classmethod
    def is_usable(cls):
        try:
            return subprocess.call(["git", "rev-parse", "--git-dir"], stderr=subprocess.PIPE, stdout=subprocess.PIPE) == 0
        except OSError as err:
            if err.errno in (errno.ENOENT, errno.EACCES):
                return False
            raise

    @classmethod
    def assert_non_dirty(cls):
        lines = [
            line.strip()
            for line in subprocess.check_output(["git", "status", "--porcelain"]).splitlines()
            if not line.strip().startswith(b"??")
        ]

        if lines:
            raise WorkingDirectoryIsDirtyException("Git working directory is not clean:\n{}".format(b"\n".join(lines).decode()))

    @classmethod
    def latest_tag_info(cls):
        try:
            # git-describe doesn't update the git-index, so we do that
            subprocess.check_output(["git", "update-index", "--refresh"])

            # get info about the latest tag in git
            describe_out = (
                subprocess.check_output(
                    [
                        "git",
                        "describe",
                        "--dirty",
                        "--tags",
                        "--long",
                        "--abbrev=40",
                        "--match=r*",
                    ],
                    stderr=subprocess.STDOUT,
                )
                .decode()
                .split("-")
            )
        except subprocess.CalledProcessError:
            logger.debug("Error when running git describe")
            return {}

        info = {}

        if describe_out[-1].strip() == "dirty":
            info["dirty"] = True
            describe_out.pop()

        info["commit_sha"] = describe_out.pop().lstrip("g")
        info["distance_to_latest_tag"] = int(describe_out.pop())
        info["current_version"] = "-".join(describe_out).lstrip("r")

        return info

    @classmethod
    def add_path(cls, path):
        subprocess.check_output(["git", "add", "--update", path])

    @classmethod
    def tag(cls, sign, name, message):
        """
        Create a tag of the new_version in Git.

        If only name is given, bumpversion uses a lightweight tag.
        Otherwise, it utilizes an annotated tag.
        """
        command = ["git", "tag", name]
        if sign:
            command += ["--sign"]
        if message:
            command += ["--message", message]
        subprocess.check_output(command)



import re


class NumericFunction:
    """
    This is a class that provides a numeric function for version parts.
    It simply starts with the provided first_value (0 by default) and increases it following the sequence of integer numbers.

    This function also supports alphanumeric parts, altering just the numeric part (e.g. 'r3' --> 'r4'). Only the first numeric group
    found in the part is considered (e.g. 'r3-001' --> 'r4-001').
    """

    FIRST_NUMERIC = re.compile(r"([^\d]*)(\d+)(.*)")

    def __init__(self, first_value=None):

        if first_value is not None:
            try:
                _, _, _ = self.FIRST_NUMERIC.search(first_value).groups()
            except AttributeError:
                # pylint: disable=raise-missing-from
                raise ValueError(f"The given first value {first_value} does not contain any digit")
        else:
            first_value = 0

        self.first_value = str(first_value)

    def bump(self, value):
        part_prefix, part_numeric, part_suffix = self.FIRST_NUMERIC.search(value).groups()
        bumped_numeric = int(part_numeric) + 1

        return "".join([part_prefix, str(bumped_numeric), part_suffix])



# pylint: skip-file

import os
import subprocess
from functools import partial
from textwrap import dedent

# noinspection PyPackageRequirements
import pytest
# noinspection PyPackageRequirements
from testfixtures import LogCapture

from bumpsemver import exceptions
from bumpsemver.cli import main
from tests.test_cli import EXPECTED_OPTIONS, check_output, COMMIT, COMMIT_NOT_TAG

check_call = partial(subprocess.check_call, env=os.environ.copy())


def test_regression_help_in_work_dir(tmpdir, capsys):
    tmpdir.chdir()
    tmpdir.join("some_source.txt").write("1.7.2013")
    check_call(["git", "init"])
    check_call(["git", "add", "some_source.txt"])
    check_call(["git", "commit", "-m", "initial commit"])
    check_call(["git", "tag", "r1.7.2013"])

    with pytest.raises(SystemExit):
        main(["--help"])

    out, err = capsys.readouterr()

    for option_line in EXPECTED_OPTIONS:
        assert option_line in out, f"Usage string is missing {option_line}"

    assert "Version that needs to be updated (default: 1.7.2013)" in out


def test_dirty_work_dir(tmpdir):
    tmpdir.chdir()
    check_call(["git", "init"])
    tmpdir.join("dirty").write("i'm dirty")

    check_call(["git", "add", "dirty"])

    with pytest.raises(exceptions.WorkingDirectoryIsDirtyException):
        with LogCapture() as log_capture:
            main(["patch", "--current-version", "1", "--new-version", "2", "file7"])

    log_capture.check_present(
        (
            "bumpsemver.cli",
            "WARNING",
            "Git working directory is not clean:\n"
            "A  dirty\n"
            "\n"
            "Use --allow-dirty to override this if you know what you're doing."
        )
    )


def test_force_dirty_work_dir(tmpdir):
    tmpdir.chdir()
    check_call(["git", "init"])
    tmpdir.join("dirty2").write("i'm dirty! 1.1.1")

    check_call(["git", "add", "dirty2"])

    main(["patch", "--allow-dirty", "--current-version", "1.1.1", "dirty2"])

    assert "i'm dirty! 1.1.2" == tmpdir.join("dirty2").read()


def test_commit_and_tag(tmpdir):
    tmpdir.chdir()
    check_call(["git", "init"])
    tmpdir.join("VERSION").write("47.1.1")
    check_call(["git", "add", "VERSION"])
    check_call(["git", "commit", "-m", "initial commit"])

    main(["patch", "--current-version", "47.1.1", "--commit", "VERSION"])

    assert "47.1.2" == tmpdir.join("VERSION").read()

    log = check_output(["git", "log", "-p"]).decode("utf-8")

    assert "-47.1.1" in log
    assert "+47.1.2" in log
    assert "[OPS] bumped version: 47.1.1 → 47.1.2" in log

    tag_out = check_output(["git", "tag"])

    assert b"r47.1.2" not in tag_out

    main(["patch", "--current-version", "47.1.2", "--commit", "--tag", "VERSION"])

    assert "47.1.3" == tmpdir.join("VERSION").read()

    check_output(["git", "log", "-p"])
    tag_out = check_output(["git", "tag"])

    assert b"r47.1.3" in tag_out


def test_commit_and_tag_with_configfile(tmpdir):
    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        commit = True
        tag = True
        """).strip())

    check_call(["git", "init"])
    tmpdir.join("VERSION").write("48.1.1")
    check_call(["git", "add", "VERSION"])
    check_call(["git", "commit", "-m", "initial commit"])

    main(["patch", "--current-version", "48.1.1", "--no-tag", "VERSION"])

    assert "48.1.2" == tmpdir.join("VERSION").read()

    log = check_output(["git", "log", "-p"]).decode("utf-8")

    assert "-48.1.1" in log
    assert "+48.1.2" in log
    assert "[OPS] bumped version: 48.1.1 → 48.1.2" in log

    tag_out = check_output(["git", "tag"])

    assert b"r48.1.2" not in tag_out

    main(["patch", "--current-version", "48.1.2", "VERSION"])

    assert "48.1.3" == tmpdir.join("VERSION").read()

    check_output(["git", "log", "-p"])

    tag_out = check_output(["git", "tag"])

    assert b"r48.1.3" in tag_out


@pytest.mark.parametrize("config", [COMMIT, COMMIT_NOT_TAG])
def test_commit_and_not_tag_with_configfile(tmpdir, config):
    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write(config)

    check_call(["git", "init"])
    tmpdir.join("VERSION").write("48.10.1")
    check_call(["git", "add", "VERSION"])
    check_call(["git", "commit", "-m", "initial commit"])

    main(["patch", "--current-version", "48.10.1", "VERSION"])

    assert "48.10.2" == tmpdir.join("VERSION").read()

    log = check_output(["git", "log", "-p"]).decode("utf-8")

    assert "-48.10.1" in log
    assert "+48.10.2" in log
    assert "[OPS] bumped version: 48.10.1 → 48.10.2" in log

    tag_out = check_output(["git", "tag"])

    assert b"v48.10.2" not in tag_out


def test_commit_explicitly_false(tmpdir):
    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version: 10.0.0
        commit = False
        tag = False
        """).strip())

    check_call(["git", "init"])
    tmpdir.join("tracked_file").write("10.0.0")
    check_call(["git", "add", "tracked_file"])
    check_call(["git", "commit", "-m", "initial commit"])

    main(["patch", "tracked_file"])

    assert "10.0.1" == tmpdir.join("tracked_file").read()

    log = check_output(["git", "log", "-p"]).decode("utf-8")
    assert "10.0.1" not in log

    diff = check_output(["git", "diff"]).decode("utf-8")
    assert "10.0.1" in diff


def test_commit_configfile_true_cli_false_override(tmpdir):
    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version: 27.0.0
        commit = True
        """).strip())

    check_call(["git", "init"])
    tmpdir.join("dont_commit_file").write("27.0.0")
    check_call(["git", "add", "dont_commit_file"])
    check_call(["git", "commit", "-m", "initial commit"])

    main(["patch", "--no-commit", "dont_commit_file"])

    assert "27.0.1" == tmpdir.join("dont_commit_file").read()

    log = check_output(["git", "log", "-p"]).decode("utf-8")
    assert "27.0.1" not in log

    diff = check_output(["git", "diff"]).decode("utf-8")
    assert "27.0.1" in diff


def test_current_version_from_tag(tmpdir):
    # prepare
    tmpdir.join("update_from_tag").write("26.6.0")
    tmpdir.chdir()
    check_call(["git", "init"])
    check_call(["git", "add", "update_from_tag"])
    check_call(["git", "commit", "-m", "initial"])
    check_call(["git", "tag", "r26.6.0"])

    # don't give current-version, that should come from tag
    main(["patch", "update_from_tag"])

    assert tmpdir.join("update_from_tag").read() == "26.6.1"


def test_current_version_from_tag_written_to_config_file(tmpdir):
    # prepare
    tmpdir.join("updated_also_in_config_file").write("14.6.0")
    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write("""[bumpsemver]""")

    check_call(["git", "init"])
    check_call(["git", "add", "updated_also_in_config_file"])
    check_call(["git", "commit", "-m", "initial"])
    check_call(["git", "tag", "r14.6.0"])

    # don't give current-version, that should come from tag
    main(["patch", "updated_also_in_config_file", "--commit", "--tag"])

    assert "14.6.1" == tmpdir.join("updated_also_in_config_file").read()
    assert "14.6.1" in tmpdir.join(".bumpsemver.cfg").read()


def test_override_vcs_current_version(tmpdir):
    # prepare
    tmpdir.join("contains_actual_version").write("6.7.8")
    tmpdir.chdir()
    check_call(["git", "init"])
    check_call(["git", "add", "contains_actual_version"])
    check_call(["git", "commit", "-m", "initial"])
    check_call(["git", "tag", "v6.7.8"])

    # update file
    tmpdir.join("contains_actual_version").write("7.0.0")
    check_call(["git", "add", "contains_actual_version"])

    # but forgot to tag or forgot to push --tags
    check_call(["git", "commit", "-m", "major release"])

    # if we don't give current-version here we get "AssertionError:
    # Did not find string 6.7.8 in file contains_actual_version"
    main(["patch", "--current-version", "7.0.0", "contains_actual_version"])

    assert "7.0.1" == tmpdir.join("contains_actual_version").read()


def test_read_version_tags_only(tmpdir):
    # prepare
    tmpdir.join("update_from_tag").write("29.6.0")
    tmpdir.chdir()
    check_call(["git", "init"])
    check_call(["git", "add", "update_from_tag"])
    check_call(["git", "commit", "-m", "initial"])
    check_call(["git", "tag", "r29.6.0"])
    check_call(["git", "commit", "--allow-empty", "-m", "a commit"])
    check_call(["git", "tag", "jenkins-deploy-my-project-2"])

    # don't give current-version, that should come from tag
    main(["patch", "update_from_tag"])

    assert "29.6.1" == tmpdir.join("update_from_tag").read()


def test_tag_name(tmpdir):
    tmpdir.chdir()
    check_call(["git", "init"])
    tmpdir.join("VERSION").write("31.1.1")
    check_call(["git", "add", "VERSION"])
    check_call(["git", "commit", "-m", "initial commit"])

    main(["patch", "--current-version", "31.1.1", "--commit", "--tag", "VERSION", "--tag-name", "ReleasedVersion-{new_version}"])

    tag_out = check_output(["git", "tag"])

    assert b"ReleasedVersion-31.1.2" in tag_out


def test_message_from_config_file(tmpdir):
    tmpdir.chdir()
    check_call(["git", "init"])
    tmpdir.join("VERSION").write("400.0.0")
    check_call(["git", "add", "VERSION"])
    check_call(["git", "commit", "-m", "initial commit"])

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version: 400.0.0
        new_version: 401.0.0
        commit: True
        tag: True
        message: {current_version} was old, {new_version} is new
        tag_name: from-{current_version}-to-{new_version}
        """).strip())

    main(["major", "VERSION"])

    log = check_output(["git", "log", "-p"])

    assert b"400.0.0 was old, 401.0.0 is new" in log

    tag_out = check_output(["git", "tag"])

    assert b"from-400.0.0-to-401.0.0" in tag_out


def test_unannotated_tag(tmpdir):
    tmpdir.chdir()
    check_call(["git", "init"])
    tmpdir.join("VERSION").write("42.3.1")
    check_call(["git", "add", "VERSION"])
    check_call(["git", "commit", "-m", "initial commit"])

    main(["patch", "--current-version", "42.3.1", "--commit", "--tag", "VERSION",
          "--tag-name", "ReleasedVersion-{new_version}", "--tag-message", ""])

    tag_out = check_output(["git", "tag"])
    assert b"ReleasedVersion-42.3.2" in tag_out

    describe_out = subprocess.call(["git", "describe"])
    assert 128 == describe_out

    describe_out = subprocess.check_output(["git", "describe", "--tags"])
    assert describe_out.startswith(b"ReleasedVersion-42.3.2")


def test_annotated_tag(tmpdir):
    tmpdir.chdir()
    check_call(["git", "init"])
    tmpdir.join("VERSION").write("42.4.1")
    check_call(["git", "add", "VERSION"])
    check_call(["git", "commit", "-m", "initial commit"])

    main(["patch", "--current-version", "42.4.1", "--commit", "--tag", "VERSION", "--tag-message", "test {new_version}-tag"])
    assert "42.4.2" == tmpdir.join("VERSION").read()

    tag_out = check_output(["git", "tag"])
    assert b"r42.4.2" in tag_out

    describe_out = subprocess.check_output(["git", "describe"])
    assert describe_out == b"r42.4.2\n"

    describe_out = subprocess.check_output(["git", "show", "r42.4.2"])
    assert describe_out.startswith(b"tag r42.4.2\n")
    assert b"test 42.4.2-tag" in describe_out


def test_vcs_describe(tmpdir):
    tmpdir.chdir()
    check_call(["git", "init"])
    tmpdir.join("VERSION").write("42.5.1")
    check_call(["git", "add", "VERSION"])
    check_call(["git", "commit", "-m", "initial commit"])

    main(["patch", "--current-version", "42.5.1", "--commit", "--tag", "VERSION", "--tag-message", "test {new_version}-tag"])
    assert "42.5.2" == tmpdir.join("VERSION").read()

    describe_out = subprocess.check_output(["git", "describe"])
    assert b"r42.5.2\n" == describe_out

    main(["patch", "--current-version", "42.5.2", "--commit", "--tag", "VERSION",
          "--tag-name", "ReleasedVersion-{new_version}", "--tag-message", ""])
    assert "42.5.3" == tmpdir.join("VERSION").read()

    describe_only_annotated_out = subprocess.check_output(["git", "describe"])
    assert describe_only_annotated_out.startswith(b"r42.5.2-1-g")

    describe_all_out = subprocess.check_output(["git", "describe", "--tags"])
    assert b"ReleasedVersion-42.5.3\n" == describe_all_out


def test_utf8_message_from_config_file(tmpdir):
    tmpdir.chdir()
    check_call(["git", "init"])
    tmpdir.join("VERSION").write("10.10.0")
    check_call(["git", "add", "VERSION"])
    check_call(["git", "commit", "-m", "initial commit"])

    initial_config = dedent("""
        [bumpsemver]
        current_version = 10.10.0
        commit = True
        message = [{now}] [{utcnow} {utcnow:%YXX%mYY%d}]
        """).strip()
    tmpdir.join(".bumpsemver.cfg").write(initial_config)

    main(['major', "VERSION"])

    log = check_output(["git", "log", "-p"])

    assert b"[20" in log
    assert b"] [" in log
    assert b"XX" in log
    assert b"YY" in log


def test_commit_and_tag_from_below_vcs_root(tmpdir, monkeypatch):
    tmpdir.chdir()
    check_call(["git", "init"])
    tmpdir.join("VERSION").write("30.0.3")
    check_call(["git", "add", "VERSION"])
    check_call(["git", "commit", "-m", "initial commit"])

    tmpdir.mkdir("subdir")
    monkeypatch.chdir(tmpdir.join("subdir"))

    main(["major", "--current-version", "30.0.3", "--commit", "../VERSION"])

    assert "31.0.0" == tmpdir.join("VERSION").read()


def test_subjunctive_dry_run_logging(tmpdir):
    tmpdir.join("dont_touch_me.txt").write("0.8.0")
    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 0.8.0
        commit = True
        tag = True
        [bumpsemver:file:dont_touch_me.txt]
    """).strip())

    check_call(["git", "init"])
    check_call(["git", "add", "dont_touch_me.txt"])
    check_call(["git", "commit", "-m", "initial commit"])

    with LogCapture() as log_capture:
        main(["patch", "--verbose", "--dry-run"])

    log_capture.check(
        ("bumpsemver.cli", "INFO", "Reading config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO",
         "[bumpsemver]\ncurrent_version = 0.8.0\ncommit = True\ntag = True\n[bumpsemver:file:dont_touch_me.txt]"),
        ("bumpsemver.version_part", "INFO", "Parsing version '0.8.0' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=0, minor=8, patch=0"),
        ("bumpsemver.cli", "INFO", "Attempting to increment part 'patch'"),
        ("bumpsemver.cli", "INFO", "Values are now: major=0, minor=8, patch=1"),
        ("bumpsemver.cli", "INFO", 'Dry run active, won\'t touch any files.'),
        ("bumpsemver.version_part", "INFO", "Parsing version '0.8.1' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=0, minor=8, patch=1"),
        ("bumpsemver.cli", "INFO", "New version will be '0.8.1'"),
        ("bumpsemver.cli", "INFO", "Asserting files dont_touch_me.txt contain the version string..."),
        ("bumpsemver.files.generic", "INFO", "Found '0.8.0' in dont_touch_me.txt at line 0: 0.8.0"),
        ("bumpsemver.files.generic", "INFO", "Would change generic file dont_touch_me.txt:"),
        ("bumpsemver.files.generic", "INFO", "--- a/dont_touch_me.txt\n+++ b/dont_touch_me.txt\n@@ -1 +1 @@\n-0.8.0\n+0.8.1"),
        ("bumpsemver.cli", "INFO", "Would write to config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO",
         "[bumpsemver]\ncurrent_version = 0.8.1\ncommit = True\ntag = True\n\n[bumpsemver:file:dont_touch_me.txt]\n\n"),
        ("bumpsemver.cli", "INFO", "Would prepare Git commit"),
        ("bumpsemver.cli", "INFO", "Would add changes in file 'dont_touch_me.txt' to Git"),
        ("bumpsemver.cli", "INFO", "Would add changes in file '.bumpsemver.cfg' to Git"),
        ("bumpsemver.cli", "INFO", "Would commit to Git with message '[OPS] bumped version: 0.8.0 \u2192 0.8.1'"),
        ("bumpsemver.cli", "INFO", "Would tag `r0.8.1` with message `[OPS] bumped version: 0.8.0 \u2192 0.8.1` in Git and not signing")
    )


def test_log_commit_message_if_no_commit_tag_but_usable_vcs(tmpdir):
    tmpdir.join("please_touch_me.txt").write("0.3.3")
    tmpdir.chdir()

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
        [bumpsemver]
        current_version = 0.3.3
        commit = False
        tag = False
        [bumpsemver:file:please_touch_me.txt]
        """).strip())

    check_call(["git", "init"])
    check_call(["git", "add", "please_touch_me.txt"])
    check_call(["git", "commit", "-m", "initial commit"])

    with LogCapture() as log_capture:
        main(["patch", "--verbose"])

    log_capture.check(
        ("bumpsemver.cli", "INFO", "Reading config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO",
         "[bumpsemver]\ncurrent_version = 0.3.3\ncommit = False\ntag = False\n[bumpsemver:file:please_touch_me.txt]"),
        ("bumpsemver.version_part", "INFO", "Parsing version '0.3.3' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=0, minor=3, patch=3"),
        ("bumpsemver.cli", "INFO", "Attempting to increment part 'patch'"),
        ("bumpsemver.cli", "INFO", "Values are now: major=0, minor=3, patch=4"),
        ("bumpsemver.version_part", "INFO", "Parsing version '0.3.4' using regexp '(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)'"),
        ("bumpsemver.version_part", "INFO", "Parsed the following values: major=0, minor=3, patch=4"),
        ("bumpsemver.cli", "INFO", "New version will be '0.3.4'"),
        ("bumpsemver.cli", "INFO", "Asserting files please_touch_me.txt contain the version string..."),
        ("bumpsemver.files.generic", "INFO", "Found '0.3.3' in please_touch_me.txt at line 0: 0.3.3"),
        ("bumpsemver.files.generic", "INFO", "Changing generic file please_touch_me.txt:"),
        ("bumpsemver.files.generic", "INFO", "--- a/please_touch_me.txt\n+++ b/please_touch_me.txt\n@@ -1 +1 @@\n-0.3.3\n+0.3.4"),
        ("bumpsemver.cli", "INFO", "Writing to config file .bumpsemver.cfg:"),
        ("bumpsemver.cli", "INFO",
         "[bumpsemver]\ncurrent_version = 0.3.4\ncommit = False\ntag = False\n\n[bumpsemver:file:please_touch_me.txt]\n\n"),
        ("bumpsemver.cli", "INFO", "Would prepare Git commit"),
        ("bumpsemver.cli", "INFO", "Would add changes in file 'please_touch_me.txt' to Git"),
        ("bumpsemver.cli", "INFO", "Would add changes in file '.bumpsemver.cfg' to Git"),
        ("bumpsemver.cli", "INFO", "Would commit to Git with message '[OPS] bumped version: 0.3.3 \u2192 0.3.4'"),
        ("bumpsemver.cli", "INFO", "Would tag `r0.3.4` with message `[OPS] bumped version: 0.3.3 \u2192 0.3.4` in Git and not signing"),
    )


def test_regression_tag_name_with_hyphens(tmpdir):
    tmpdir.chdir()
    tmpdir.join("some_source.txt").write("2014.10.22")
    check_call(["git", "init"])
    check_call(["git", "add", "some_source.txt"])
    check_call(["git", "commit", "-m", "initial commit"])
    check_call(["git", "tag", "very-unrelated-but-containing-lots-of-hyphens"])

    tmpdir.join(".bumpsemver.cfg").write(dedent("""
    [bumpsemver]
    current_version = 2014.10.22
    """))

    main(["patch", 'some_source.txt'])


def test_unclean_repo_exception(tmpdir, caplog):
    tmpdir.chdir()

    config = dedent("""
        [bumpsemver]
        current_version = 0.0.0
        tag = True
        commit = True
        message = XXX
        """).strip()

    tmpdir.join("file1").write("foo")

    # If I have a repo with an initial commit
    check_call(["git", "init"])
    check_call(["git", "add", "file1"])
    check_call(["git", "commit", "-m", "initial commit"])

    # If I add the bumpsemver config, uncommitted
    tmpdir.join(".bumpsemver.cfg").write(config)

    # I expect bumpsemver patch to fail
    with pytest.raises(subprocess.CalledProcessError):
        main(["patch"])

    # And return the output of the failing command
    assert "Failed to run" in caplog.text


def test_dry_run(tmpdir):
    tmpdir.chdir()

    config = dedent("""
        [bumpsemver]
        current_version = 0.12.0
        tag = True
        commit = True
        message = DO NOT BUMP VERSIONS WITH THIS FILE
        [bumpsemver:file:file4]
        """).strip()

    version = "0.12.0"

    tmpdir.join("file4").write(version)
    tmpdir.join(".bumpsemver.cfg").write(config)

    check_call(["git", "init"])
    check_call(["git", "add", "file4"])
    check_call(["git", "add", ".bumpsemver.cfg"])
    check_call(["git", "commit", "-m", "initial commit"])

    main(["patch", "--dry-run"])

    assert config == tmpdir.join(".bumpsemver.cfg").read()
    assert version == tmpdir.join("file4").read()

    vcs_log = check_output(["git", "log"]).decode('utf-8')

    assert "initial commit" in vcs_log
    assert "DO NOT" not in vcs_log
